// Device-specific include file.
#include <p33fxxxx.h>

// C functions.
#include <string.h>

// VDC functions.
#include "vdc.h"
#include "vdc_graphics.h"
#include "vdc_text.h"
#include "vdc_text_font.h"

// BBC Micro VDU equates.
#include "bbc_micro_vdu.h"

// Stores the current graphics foreground and background colours.
unsigned char vdc_graphics_foreground_colour = 127;
unsigned char vdc_graphics_background_colour = 0;

// Stores the graphics cursor and origin.
struct vdc_point_2d_t vdc_graphics_cursor[3];
struct vdc_point_2d_t vdc_graphics_origin;

// Stores the current VDU text mode.
enum bbc_micro_vdu_text_mode_t bbc_micro_vdu_text_mode = BBC_MICRO_VDU_TEXT_MODE_TEXT;

unsigned char bbc_micro_vdu_command;
unsigned char bbc_micro_vdu_command_arguments[9]; // VDU 23 is the longest, with 9 arguments.
unsigned char bbc_micro_vdu_command_argument_count = 0;
unsigned char bbc_micro_vdu_command_argument_index = 0;

// Stores whether the blinking cursor is visible or not.
volatile int bbc_micro_vdu_cursor_blink;

// Resets the emulated BBC Micro VDU.
void bbc_micro_reset(void) {
	bbc_micro_vdu_command_argument_count = 0;
	bbc_micro_vdu_command_argument_index = 0;
	bbc_micro_vdu(26);
	bbc_micro_vdu(20);
	bbc_micro_vdu(4);
	TMR3 = 0;
	PR3 = 65535;
	T3CONbits.TCKPS = 0b11; // 1:256 prescaler
	IFS0bits.T3IF = 0;
	IEC0bits.T3IE = 1;
	T3CONbits.TON = 1;
	bbc_micro_vdu_cursor_blink = !0;
}

// Write a byte to the emulated BBC Micro VDU.
void bbc_micro_vdu(unsigned char value) {
	// Reset the blink counter.
	TMR3 = 0;
	bbc_micro_vdu_cursor_blink = !0;
	// Are we waiting for an argument?
	if (bbc_micro_vdu_command_argument_index < bbc_micro_vdu_command_argument_count) {
		// Store the argument value.
		bbc_micro_vdu_command_arguments[bbc_micro_vdu_command_argument_index++] = value;
		// Have we received all arguments?
		if (bbc_micro_vdu_command_argument_index == bbc_micro_vdu_command_argument_count) {
			bbc_micro_vdu_command_argument_count = 0;
			bbc_micro_vdu_command_argument_index = 0;
			switch (bbc_micro_vdu_command) {
				case 17:
					// COLOUR
					if (bbc_micro_vdu_command_arguments[0] < 128) {
						vdc_text_foreground_colour = bbc_micro_vdu_command_arguments[0];
					} else {
						vdc_text_background_colour = bbc_micro_vdu_command_arguments[0] & 127;
					}
					break;
				case 18:
					// GCOL
					vdc_graphics_plot_mode = bbc_micro_vdu_command_arguments[0] % 5;
					if (bbc_micro_vdu_command_arguments[1] < 128) {
						vdc_graphics_foreground_colour = bbc_micro_vdu_command_arguments[1];
					} else {
						vdc_graphics_background_colour = bbc_micro_vdu_command_arguments[1] & 127;
					}
					break;
				case 22:
					// MODE
					bbc_micro_vdu(4);
					bbc_micro_vdu(20);
					bbc_micro_vdu(26);
					vdc_text_set_font_size(bbc_micro_vdu_command_arguments[0] ? VDC_TEXT_FONT_SIZE_LARGE : VDC_TEXT_FONT_SIZE_SMALL);
					break;
				case 23:
					// TODO: VDU 23
					break;
				case 24:
					// TODO: Graphics viewport.
					break;
				case 25:
					// TODO: PLOT
					break;
				case 27:
					// Escaped character.
					vdc_text_put_char(bbc_micro_vdu_command_arguments[0]);
					break;
				case 28:
					// Text viewport.
					{
						struct vdc_rectangle_t new_text_viewport = {
							bbc_micro_vdu_command_arguments[0],
							bbc_micro_vdu_command_arguments[3],
							bbc_micro_vdu_command_arguments[2],
							bbc_micro_vdu_command_arguments[1]
						};
						vdc_text_set_viewport(new_text_viewport);
					}
					break;
				case 29:
					// TODO: Graphics origin.
					break;
				case 31:
					// TAB
					if (bbc_micro_vdu_command_arguments[0] >= vdc_text_viewport.left && bbc_micro_vdu_command_arguments[0] <= vdc_text_viewport.right) vdc_text_cursor_position.x = bbc_micro_vdu_command_arguments[0];
					if (bbc_micro_vdu_command_arguments[1] >= vdc_text_viewport.top && bbc_micro_vdu_command_arguments[1] <= vdc_text_viewport.bottom) vdc_text_cursor_position.y = bbc_micro_vdu_command_arguments[1];
					break;
			}
		}
	} else {
		bbc_micro_vdu_command = value;
		if (value < 32) {
			switch (value) {
				case 0:
					// Do nothing.
					break;
				case 4:
					bbc_micro_vdu_text_mode = BBC_MICRO_VDU_TEXT_MODE_TEXT;
					break;
				case 5:
					bbc_micro_vdu_text_mode = BBC_MICRO_VDU_TEXT_MODE_GRAPHICS;
					break;
				case 8:
					vdc_text_move_cursor_left();
					break;
				case 9:
					vdc_text_move_cursor_right();
					break;
				case 10:
					vdc_text_move_cursor_down();
					break;
				case 11:
					vdc_text_move_cursor_right();
					break;
				case 12:
					// CLS
					vdc_text_clear();
					break;
				case 13:
					vdc_text_cursor_position.x = vdc_text_viewport.left;
					break;
				case 14:
					// TODO: Enable auto-paging mode.
					break;
				case 15:
					// TODO: Disable auto-paging mode.
					break;
				case 16:
					vdc_graphics_set_colour_and_plot_mode(vdc_graphics_background_colour, vdc_graphics_plot_mode);
					vdc_graphics_clear();
					break;
				case 17:
					// COLOUR
					bbc_micro_vdu_command_argument_count = 1;
					break;
				case 18:
					// GCOL
					bbc_micro_vdu_command_argument_count = 2;
					break;
				case 20:
					vdc_graphics_background_colour = vdc_text_background_colour = 0;
					vdc_graphics_foreground_colour = vdc_text_foreground_colour = 127;
					vdc_graphics_plot_mode = VDC_GRAPHICS_PLOT_PLOT;
					break;
				case 22:
					// MODE
					bbc_micro_vdu_command_argument_count = 1;
					break;
				case 23:
					// VDU 23
					bbc_micro_vdu_command_argument_count = 9;
					break;
				case 24:
					// Graphics viewport.
					bbc_micro_vdu_command_argument_count = 8;
					break;
				case 25:
					// PLOT
					bbc_micro_vdu_command_argument_count = 5;
					break;
				case 26:
					// Reset the text viewport and cursor.
					vdc_text_cursor_position.x = 0;
					vdc_text_cursor_position.y = 0;
					vdc_text_reset_viewport();
					// Reset the graphics viewport, cursor and origin.
					vdc_graphics_viewport.left = 0;
					vdc_graphics_viewport.top = 0;
					vdc_graphics_viewport.right = 319;
					vdc_graphics_viewport.bottom = 239;
					memset(vdc_graphics_cursor, 0, sizeof(vdc_graphics_cursor));
					vdc_graphics_origin.x = 0;
					vdc_graphics_origin.y = 0;
					break;
				case 27:
					// Escape
					bbc_micro_vdu_command_argument_count = 1;
					break;
				case 28:
					// Text viewport.
					bbc_micro_vdu_command_argument_count = 4;
					break;
				case 29:
					// Graphics origin.
					bbc_micro_vdu_command_argument_count = 4;
					break;
				case 30:
					switch (bbc_micro_vdu_text_mode) {
						case BBC_MICRO_VDU_TEXT_MODE_TEXT:
							// Home the text cursor.
							vdc_text_cursor_position.x = vdc_text_viewport.left;
							vdc_text_cursor_position.y = vdc_text_viewport.top;
							break;
						case BBC_MICRO_VDU_TEXT_MODE_GRAPHICS:
							// TODO: Home the graphics cursor.
							break;
					}
					break;
				case 31:
					// TAB
					bbc_micro_vdu_command_argument_count = 2;
					break;
			}
		} else if (value == 127) {
			// VDU 127 deletes the character to the left of the cursor and backspaces the cursor to this position.
			vdc_text_move_cursor_left();
			vdc_text_put_map(' ');
		} else {
			// Output the character directly to the screen.
			vdc_text_put_char(value);
		}
	}
}

int bbc_micro_vdu_cursor_shown = 0;
unsigned int bbc_micro_vdu_under_cursor[8];
unsigned int* bbc_micro_vdu_under_cursor_address;
// Show the cursor on the screen.
void bbc_micro_vdu_cursor_show(void) {
	// If the cursor is already shown, bail out.
	if (bbc_micro_vdu_cursor_shown) return;
	// Copy the content under the cursor to the temporary buffer.
	bbc_micro_vdu_under_cursor_address = vdc_frame_buffer + (vdc_text_cursor_position.y * 20 * 8) + (vdc_text_cursor_position.x / (16 / vdc_text_font_size));
	unsigned int* s = bbc_micro_vdu_under_cursor_address;
	unsigned int d;
	for (d = 0; d < 8; ++d, s += 20) {
		bbc_micro_vdu_under_cursor[d] = *s;
	}
	// Draw the cursor.
	vdc_text_put_map('_');
	// The cursor has been shown.
	bbc_micro_vdu_cursor_shown = !0;
}


// Remove the cursor from the screen.
void bbc_micro_vdu_cursor_hide(void) {
	// If the cursor is already hidden, bail out.
	if (!bbc_micro_vdu_cursor_shown) return;
	// Copy the content under the cursor back from the temporary buffer.
	unsigned int* d = bbc_micro_vdu_under_cursor_address;
	unsigned int s;
	for (s = 0; s < 8; ++s, d += 20) {
		*d = bbc_micro_vdu_under_cursor[s];
	}
	// The cursor has been hidden.
	bbc_micro_vdu_cursor_shown = 0;
}

void __attribute__((interrupt, auto_psv)) _T3Interrupt(void) {
	bbc_micro_vdu_cursor_blink ^= !0;
	IFS0bits.T3IF = 0;
	IEC0bits.T3IE = 1;
}

// Draw a blinking cursor (call in a loop).
void bbc_micro_vdu_cursor_draw_blinking(void) {
	if (bbc_micro_vdu_cursor_blink) {
		bbc_micro_vdu_cursor_show();
	} else {
		bbc_micro_vdu_cursor_hide();
	}
}
